-- Coded by Indev for Sunflower's Garden
-- Not using type constants directly since modded things may not be loaded yet (when they're added to the list)
local mobjnames = {
    -- Daytona trees
    "MT_PINETREE",

    -- Sky Garden vines
    "MT_SGVINE1", "MT_SGVINE2", "MT_SGVINE3",

    -- Peach Gardens trees
    "MT_PGTREE",

    -- Opulence palmtree
    "MT_OPULENCE_PALMTREE",

    -- Dark race fruits
    "MT_BLUEFRUIT", "MT_ORANGEFRUIT", "MT_REDFRUIT", "MT_PINKFRUIT",

    -- SNOWMAAAAAAN
    "MT_SNOWMAN",
}

local types

-- Just so we don't need to add any hooks at runtime. This is just list of "destroyed" mobjs that should flash with MF2_DONTDRAW
local flashmobjs = {}

local function addFlashingMobj(mo)
    table.insert(flashmobjs, mo)
end

local function getMobjType(name)
    local ok, type = pcall(function() return _G[name] end)

    if not ok then
        print("\130WARNING:\128 couldn't find mobj type called "..name)

        type = -1
    end

    return type
end

local function generateTypesTable()
    types = {}

    for i = #mobjnames, 1, -1 do
        types[getMobjType(mobjnames[i])] = true
    end
end

local function isDestructible(type)
    if types == nil then
        generateTypesTable()
    end

    return types[type]
end

local function destroyObject(mo, pmo)
    mo.flags = ($ & ~(MF_SOLID|MF_NOGRAVITY)) | (MF_NOCLIP|MF_NOCLIPTHING|MF_NOCLIPHEIGHT)
    mo.fuse = TICRATE
    addFlashingMobj(mo)
    S_StartSound(mo, sfx_s1ac)

    P_InstaThrust(mo, R_PointToAngle2(0, 0, pmo.momx, pmo.momy), 3*pmo.player.speed/2)
    P_SetObjectMomZ(mo, pmo.player.speed/2)
end

local function toggleDrawFlag(mo, draw)
    if draw then
        mo.flags2 = $ & ~MF2_DONTDRAW
    else
        mo.flags2 = $ | MF2_DONTDRAW
    end
end

local function playerCollide(pmo, mo)
    if pmo.z + pmo.height < mo.z or mo.z + mo.height < pmo.z then return end

    local p = pmo.player

    if not p then return end

    if not (p.kartstuff[k_invincibilitytimer] > 0 or p.kartstuff[k_growshrinktimer] > 0) then
        return
    end

    if isDestructible(mo.type) then
        destroyObject(mo, pmo)
        return false
    end
end

addHook("MobjCollide", playerCollide, MT_PLAYER)
addHook("MobjMoveCollide", playerCollide, MT_PLAYER)

addHook("NetVars", function(net)
    types = net($)
    flashmobjs = net($) -- Just in case...
end)

addHook("ThinkFrame", function()
    local draw = leveltime % 2 == 0

    for i = #flashmobjs, 1, -1 do
        if flashmobjs[i].valid then
            toggleDrawFlag(flashmobjs[i], draw)
        else
            table.remove(flashmobjs, i)
        end
    end
end)

---------------------------------------------------------
-- Some specific mobj stuff that needs to be done
---------------------------------------------------------

---------------------------------------------------------
-- Daytona pinetrees
---------------------------------------------------------

addHook("MobjSpawn", function(mobj)
    mobj.sides = {}

    local diff = FixedAngle((360/mobj.info.mass)*FRACUNIT)

    for i = 0, mobj.info.mass do
        local ang = i * diff
        local side = P_SpawnMobj(mobj.x + cos(ang), mobj.y + sin(ang), mobj.z, MT_PINETREE_SIDE)
        side.angle = ang
        side.target = mobj
        side.threshold = i
        mobj.sides[i] = side
    end

    return true
end, MT_PINETREE)

addHook("MobjThinker", function(mobj)
    if not (mobj.valid and mobj.sides) then return end
    if mobj.momx == 0 and mobj.momy == 0 and mobj.momz == 0 then return end

    local draw = mobj.flags2 & MF2_DONTDRAW

    local diff = FixedAngle((360/mobj.info.mass)*FRACUNIT)

    for i = 0, #mobj.sides do
        local side = mobj.sides[i]

        if not side and side.valid then continue end

        local ang = i * diff

        P_MoveOrigin(side, mobj.x + cos(ang), mobj.y + sin(ang), mobj.z)

        -- Just from parent
        toggleDrawFlag(side, draw)
    end
end, MT_PINETREE)

addHook("MobjFuse", function(mobj)
    if not (mobj.valid and mobj.sides) then return end

    for i = 0, #mobj.sides do
        local side = mobj.sides[i]

        if not side and side.valid then continue end

        P_RemoveMobj(side)
    end
end, MT_PINETREE)

---------------------------------------------------------
-- Opulence palmtrees
---------------------------------------------------------

-- I'm sorry perfomance :(
-- But i want to do the funny tree kick, not have tree flashing and frozen in place
-- At least hopefully MF_SCENERY compensates that
mobjinfo[MT_OPULENCE_PALMTREE].flags = ($ & ~MF_NOTHINK) | MF_SCENERY

---------------------------------------------------------
-- Snowman
---------------------------------------------------------

addHook("MobjThinker", function(mobj)
    if mobj.valid and mobj.fuse == 2 then
        -- "..."
        -- "I BETTER NOT SEE YOU AGAIN."
        P_RemoveMobj(mobj)
        return
    end
end, MT_SNOWMAN)
